VERSION = (0, 9, 8)

__title__ = "localzone"
__description__ = "A simple library for managing DNS zones."
__url__ = "https://localzone.iomaestro.com"
__version__ = ".".join(map(str, VERSION))
__author__ = "Andrew Grant Spencer"
__author_email__ = "ags@iomaestro.com"
__license__ = "BSD"
__copyright__ = "Copyright (C) 2018 Andrew Grant Spencer"
